﻿/*
 This file is part of ELoadV1.

    ELoadV1 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ELoadV1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ELoadV1.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO.Ports;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Windows.Forms.DataVisualization.Charting;

[StructLayout(LayoutKind.Explicit, Size = 64)]
public unsafe struct DataConverter
{
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 64)]
    [FieldOffset(0)]
    public fixed byte b[64];
    [FieldOffset(0)]
    public fixed float f[64 / 4];
    [FieldOffset(0)]
    public fixed UInt16 ui16[64 / 2];
}

namespace ELoadV1
{

    public partial class Form1 : Form
    {
        const string project_name = "Electronic Load";
        const string project_version = " V 1.1";

        public const UInt16 tx_size = 24, rx_size = 24, tx_size_conf = 52;
        public const int bufsize = 30;

        const byte ERR_NO_ERR = 0;
        const byte ERR_LIM_OVERCURR = 1;
        const byte ERR_LIM_OVERVOLT = 2;
        const byte ERR_LIM_OVERPOW = 4;
        const byte ERR_LIM_UVLO = 8;
        const byte ERR_LIM_OVERTEMP = 16;
        const byte ERR_COM_TIMEOUT = 32;
        const byte ERR_COM_CHECKSUM = 64;
        const byte ERR_NO_INIT = 128;

        public static Int32 interval_cycle;
        public static UInt16 recCnt = 0;
        public static DataConverter recConv;

        public static bool send_now = false;
        public static bool rx_ok = false, checksum_fail = false;


        public struct LoadDataSet
        {
            public float curr;
            public float voltage;
            public float resistance;
            public float power;
            public float uvlo;
            public byte mode;
            public byte ack;
        };
        public struct LoadDataAct
        {
            public float curr;
            public float voltage;
            public float resistance;
            public float power;
            public float temp;
            public byte mode;
            public byte status;
        };

        public struct LoadDataStruct
        {
            public LoadDataAct act;
            public LoadDataSet set;
        };

        public static LoadDataStruct LoadData;

        public struct configData
        {
            public float max_power;
            public float max_voltage;
            public float max_current;
            public float max_temperature;
            public float current_input_scale;
            public float current_output_scale;
            public float voltage_input_scale;
            public float current_controller_kI;
            public float current_controller_limit;
            public float voltage_controller_kI;
            public float voltage_controller_kP;
            public float cable_resistance;
            public bool configuration_loaded;
            public bool configuration_written;
        };


        public static Thread comm_thread;
        public static bool thread_active = false;
        public static bool check_timeout = false;


        DataLogger logger = new DataLogger();

        static configData conf;

        public int chart_data_cnt;
        public UInt32 chart_data_counter = 0;
        Series series_act_current = new Series("Act Curr.");
        Series series_set_current = new Series("Set Curr.");
        Series series_act_power = new Series("Act Pwr.");
        Series series_set_power = new Series("Set Pwr.");
        Series series_act_volt = new Series("Act Volt.");
        Series series_set_volt = new Series("Set Volt.");


        public Form1()
        {
            InitializeComponent();
            FormClosing += new FormClosingEventHandler(mainWindow_FormClosing);
            
            this.Text = project_name + project_version;

            ConfigureLoadDataStorage();
            initGUI();
            //Serial
            serial.DataReceived += new SerialDataReceivedEventHandler(DataReceivedHandler);
            interval_cycle = Convert.ToInt32(txt_comm_cycle.Text);
            
            LoadData.set.mode = 0;

            chart_data_cnt = (int)(1000*Convert.ToSingle(txt_chart_time.Text) / (float)timer_chart.Interval);

            series_act_current.ChartType = SeriesChartType.FastLine;
            series_act_current.Color = Color.Red;
            series_set_current.ChartType = SeriesChartType.FastLine;
            series_set_current.Color = Color.Blue;
            chart_curr.Series.Clear();
            chart_curr.Series.Add(series_act_current);
            chart_curr.Series.Add(series_set_current);
            series_act_power.ChartType = SeriesChartType.FastLine;
            series_act_power.Color = Color.Red;
            series_set_power.ChartType = SeriesChartType.FastLine;
            series_set_power.Color = Color.Blue;
            chart_pwr.Series.Clear();
            chart_pwr.Series.Add(series_act_power);
            chart_pwr.Series.Add(series_set_power);
            series_act_volt.ChartType = SeriesChartType.FastLine;
            series_act_volt.Color = Color.Red;
            series_set_volt.ChartType = SeriesChartType.FastLine;
            series_set_volt.Color = Color.Blue;
            chart_volt.Series.Clear();
            chart_volt.Series.Add(series_act_volt);
            chart_volt.Series.Add(series_set_volt);

            chart_curr.ChartAreas[0].AxisY.IsStartedFromZero = false;
            chart_pwr.ChartAreas[0].AxisY.IsStartedFromZero = false;
            chart_volt.ChartAreas[0].AxisY.IsStartedFromZero = false;

            chart_curr.BackColor = Form1.DefaultBackColor;
            chart_volt.BackColor = Form1.DefaultBackColor;
            chart_pwr.BackColor = Form1.DefaultBackColor;

            //Config file
            try
            {
                ReadConfigData();
                
            }
            catch (Exception ex)
            {

                AppendToLog("ERR", ex.Message);
                
            }
            if (conf.configuration_loaded == false)
            {
                try
                {
                    InitConfigData();
                    WriteStdConfigData();
                    AppendToLog("ERR", "No or bad config file found.");
                    AppendToLog("INF", "Generated a sample config file. Adjust the file and restart the application.");
                }
                catch (Exception ex)
                {

                    AppendToLog("ERR", ex.Message);
                }
            }
            else
            {
                AppendToLog("INF", "Loaded a correct config file.");
            }
            

            
            
        }
        /***************************************
        ****************************************
        * GUI
        ****************************************
        ***************************************/
        void initGUI()
        {
            timer_GUI.Start();
            //Commm
            btn_comm_connect.Enabled = true;
            btn_comm_disconnect.Enabled = false;
            //Read available ports
            string[] ports = SerialPort.GetPortNames();
            foreach (string port in ports)
            {
                cb_comm_port.Items.Add(port);
            }
            if (ports.LongCount() == 0)
            {
                AppendToLog("ERR", "Found no Serial Ports! Restart the Application!");
            }
            else
            {
                AppendToLog("INF", "Found " + ports.LongCount() + "Serial Port(s).");
                cb_comm_port.SelectedIndex = 0;
            }
            cb_logging_enable.Checked = true;
            cb_comm_port.DropDownStyle = ComboBoxStyle.DropDownList;
            //Setvalues
            cB_set_Mode.SelectedItem = cB_set_Mode.SelectedIndex = 0;
            cB_set_Mode.DropDownStyle = ComboBoxStyle.DropDownList;
            txt_set_curr.BackColor = Color.LightGray;
            txt_set_power.BackColor = Color.LightGray;
            txt_set_resistance.BackColor = Color.LightGray;
            txt_set_voltage.BackColor = Color.LightGray;

            txt_set_curr.Text = String.Format("{0:0.0000}", LoadData.set.curr);
            txt_set_power.Text = String.Format("{0:00.000}", LoadData.set.power);
            txt_set_resistance.Text = String.Format("{0:000.00}", LoadData.set.resistance);
            txt_set_voltage.Text = String.Format("{0:00.000}", LoadData.set.voltage);
            txt_set_UVLO.Text = String.Format("{0:00.000}", LoadData.set.uvlo);

        }

        private void cB_set_Mode_SelectedIndexChanged(object sender, EventArgs e)
        {
            txt_set_curr.BackColor = Color.LightGray;
            txt_set_power.BackColor = Color.LightGray;
            txt_set_resistance.BackColor = Color.LightGray;
            txt_set_voltage.BackColor = Color.LightGray;
        }

        private void btn_comm_connect_Click(object sender, EventArgs e)
        {
            try
            {
                serial.BaudRate = (Int32)Convert.ToUInt32(txt_comm_baud.Text);
                serial.PortName = cb_comm_port.Text;
                serial.Open();
                btn_comm_connect.Enabled = false;
                btn_comm_disconnect.Enabled = true;
                btn_comm_connect.BackColor = Color.LightGreen;
                cb_comm_port.Enabled = false;
                txt_comm_baud.Enabled = false;
                AppendToLog("INF", "Communication established.");
                thread_active = true;
                comm_thread = new Thread(ThreadCycle);
                comm_thread.Start();
                timer_comm_timeout.Start();
                send_now = true;
                timer_chart.Start();
                logger.Clear();
            }
            catch (Exception ex)
            {
                AppendToLog("ERR",ex.Message);
            }
            
        }

        private void btn_comm_disconnect_Click(object sender, EventArgs e)
        {
            try
            {
                serial.Close();
                thread_active = false;
                btn_comm_connect.Enabled = true;
                btn_comm_disconnect.Enabled = false;
                btn_comm_connect.BackColor = btn_comm_disconnect.BackColor;
                cb_comm_port.Enabled = true;
                txt_comm_baud.Enabled = true;
                timer_chart.Stop();
                AppendToLog("INF", "Communication closed.");
            }
            catch (Exception ex)
            {
                AppendToLog("ERR", ex.Message);
            }
        }

        private void btn_comm_apply_Click(object sender, EventArgs e)
        {
            int tmp = interval_cycle;

            try
            {
                tmp = (int)Convert.ToUInt32(txt_comm_cycle.Text);
            }
            catch (Exception ex)
            {
                AppendToLog("ERR", ex.Message);
                txt_comm_cycle.Text = interval_cycle.ToString();
            }

            if (tmp >= 5 && tmp <= 1000)
                interval_cycle = tmp;
            else
            {
                AppendToLog("ERR", "Invalid value!");
                txt_comm_cycle.Text = interval_cycle.ToString();
            }
        }
        
        private void timer_chart_Tick(object sender, EventArgs e)
        {

            
            chart_data_counter++;
            while (series_act_current.Points.Count > chart_data_cnt)
            {
                series_act_current.Points.RemoveAt(0);
            }
            while (series_set_current.Points.Count > chart_data_cnt)
            {
                series_set_current.Points.RemoveAt(0);
            }
            while (series_act_power.Points.Count > chart_data_cnt)
            {
                series_act_power.Points.RemoveAt(0);
            }
            while (series_set_power.Points.Count > chart_data_cnt)
            {
                series_set_power.Points.RemoveAt(0);
            }
            while (series_act_volt.Points.Count > chart_data_cnt)
            {
                series_act_volt.Points.RemoveAt(0);
            }
            while (series_set_volt.Points.Count > chart_data_cnt)
            {
                series_set_volt.Points.RemoveAt(0);
            }
            series_act_current.Points.AddXY(chart_data_counter, LoadData.act.curr);
            if (LoadData.set.mode == 1)
            {
                series_set_current.Points.AddXY(chart_data_counter, LoadData.set.curr);
            }
            else
            {
                series_set_current.Points.Clear();
            }
            series_act_volt.Points.AddXY(chart_data_counter, LoadData.act.voltage);
            if (LoadData.set.mode == 2)
            {
                series_set_volt.Points.AddXY(chart_data_counter, LoadData.set.voltage);
            }
            else
            {
                series_set_volt.Points.Clear();
            }
            series_act_power.Points.AddXY(chart_data_counter, LoadData.act.power);
            if (LoadData.set.mode == 3)
            {
                series_set_power.Points.AddXY(chart_data_counter, LoadData.set.power);
            }
            else
            {
                series_set_power.Points.Clear();
            }
            chart_curr.ChartAreas[0].RecalculateAxesScale();
            chart_volt.ChartAreas[0].RecalculateAxesScale();
            chart_pwr.ChartAreas[0].RecalculateAxesScale();

        }

        private void btn_chart_apply_Click(object sender, EventArgs e)
        {
            float tmp;
            try
            {
                tmp =Convert.ToSingle(txt_chart_time.Text);
                if (tmp >= 1.0f)
                    chart_data_cnt = (int)(1000.0f*tmp / (float)timer_chart.Interval);
                else
                    AppendToLog("ERR", "Invalid value");
            }
            catch (Exception ex)
            {
                AppendToLog("ERR", ex.Message);
                throw;
            }
        }

        private void btn_logging_save_Click(object sender, EventArgs e)
        {
            if (logger.GetMeasLines() > 0)
            {
                try
                {
                    string logPath = System.IO.Directory.GetCurrentDirectory() + "\\";
                    string filename = DateTime.Now.ToShortDateString() + "T" + DateTime.Now.ToLongTimeString();
                    filename = filename.Replace(":", "-");
                    filename = filename.Replace(".", "-");
                    filename += ".txt";
                    filename = txt_logging_prefix.Text + filename;
                    logger.WriteLogFile(logPath + filename);
                    AppendToLog("INF", "Logfile saved. (Path: " + logPath + filename + ", " + logger.GetMeasLines().ToString() + " measurements.)");
                }
                catch (Exception ex)
                {
                    AppendToLog("ERR", ex.Message);
                    throw;
                }
            }
            else
            {
                AppendToLog("INF", "Empty file not saved.");
            }

        }

        private void btn_logging_clear_Click(object sender, EventArgs e)
        {
            try
            {
                logger.Clear();
                AppendToLog("INF", "Logfile cleared.");
            }
            catch (Exception ex)
            {
                AppendToLog("ERR", ex.Message);
                throw;
            }
        }

        private void cb_logging_enable_CheckedChanged(object sender, EventArgs e)
        {
            if (cb_logging_enable.Checked)
            {
                AppendToLog("INF", "Logging enabled.");
            }
            else
            {
                AppendToLog("INF", "Logging disabled.");
            }
        }

        private void btn_set_off_Click(object sender, EventArgs e)
        {
            cB_set_Mode.SelectedIndex = 0;
            LoadData.set.mode = 0;
        }

        private void Form1_Load(object sender, EventArgs e)
        {

        }

        private void txt_comm_cycle_TextChanged(object sender, EventArgs e)
        {
        }

        private void btn_set_apply_Click(object sender, EventArgs e)
        {
            try
            {
                LoadData.set.curr = Convert.ToSingle(txt_set_curr.Text);
                LoadData.set.power = Convert.ToSingle(txt_set_power.Text);
                LoadData.set.resistance = Convert.ToSingle(txt_set_resistance.Text);
                LoadData.set.voltage = Convert.ToSingle(txt_set_voltage.Text);
                LoadData.set.uvlo = Convert.ToSingle(txt_set_UVLO.Text);

                txt_set_curr.Text = String.Format("{0:0.0000}", LoadData.set.curr);
                txt_set_power.Text = String.Format("{0:00.000}", LoadData.set.power);
                txt_set_resistance.Text = String.Format("{0:000.00}", LoadData.set.resistance);
                txt_set_voltage.Text = String.Format("{0:00.000}", LoadData.set.voltage);
                txt_set_UVLO.Text = String.Format("{0:00.000}", LoadData.set.uvlo);

                if (cB_set_Mode.Text == "Off")
                {
                    LoadData.set.mode = 0;
                }
                else if (cB_set_Mode.Text == "Current")
                {
                    txt_set_curr.BackColor = Color.White;
                    LoadData.set.mode = 1;
                }
                else if (cB_set_Mode.Text == "Power")
                {
                    txt_set_power.BackColor = Color.White;
                    LoadData.set.mode = 3;
                }
                else if (cB_set_Mode.Text == "Voltage")
                {
                    txt_set_voltage.BackColor = Color.White;
                    LoadData.set.mode = 2;
                }
                else if (cB_set_Mode.Text == "Resistance")
                {
                    txt_set_resistance.BackColor = Color.White;
                    LoadData.set.mode = 4;
                }
                else
                {
                    //Off
                    LoadData.set.mode = 0;
                }
            }
            catch (Exception ex)
            {
                AppendToLog("ERR", ex.Message);
                txt_set_curr.Text = String.Format("{0:0.0000}", LoadData.set.curr);
                txt_set_power.Text = String.Format("{0:00.000}", LoadData.set.power);
                txt_set_resistance.Text = String.Format("{0:000.00}", LoadData.set.resistance);
                txt_set_voltage.Text = String.Format("{0:00.000}", LoadData.set.voltage);
                txt_set_UVLO.Text = String.Format("{0:00.000}", LoadData.set.uvlo);
            }

        }

        private void btn_set_quitt_Click(object sender, EventArgs e)
        {
            LoadData.set.ack = 0xFF;
        }

        public void ThreadCycle()
        {
            while (thread_active)
            {
                CommunicationCycle();
                Thread.Sleep(interval_cycle);
            }
        }

        public void CommunicationCycle()
        {
            DataConverter conv;
            byte[] sendbytes = new byte[tx_size_conf];

            if (send_now && conf.configuration_written == false)
            {
                /*  float-index Value
                 *  1	Max Power
                    2	Max Voltage
                    3	Max Current
                    4	Max Temp
                    5	Input current scale
                    6	Output current scale
                    7	Input voltage scale
                    8	Current integrator factor
                    9	Current ctrl limits
                    10	Voltage ctrl kP
                    11	Voltage ctrl kI
                    12	Cable resistance
                 */
                unsafe
                {
                    conv.b[0] = 0x00; //checksum start value
                    conv.b[1] = 0xFF; //init data

                    conv.b[2] = 0x00;//No mode
                    conv.b[3] = 0x00;//no acknowledge
                    conv.f[1] = conf.max_power;
                    conv.f[2] = conf.max_voltage;
                    conv.f[3] = conf.max_current;
                    conv.f[4] = conf.max_temperature;
                    conv.f[5] = conf.current_input_scale;
                    conv.f[6] = conf.current_output_scale;
                    conv.f[7] = conf.voltage_input_scale;
                    conv.f[8] = conf.current_controller_kI;
                    conv.f[9] = conf.current_controller_limit;
                    conv.f[10] = conf.voltage_controller_kP;
                    conv.f[11] = conf.voltage_controller_kI;
                    conv.f[12] = conf.cable_resistance;
                    for (UInt16 i = 2; i < tx_size_conf; i++)
                    {
                        conv.b[0] ^= conv.b[i];
                    }
                    //Copy the bytes to send them
                    for (UInt16 i = 0; i < tx_size_conf; i++)
                    {
                        sendbytes[i] = conv.b[i];
                    }
                }
                //Throw RX-Buffer away
                recCnt = 0;
                while (serial.BytesToRead > 0)
                {
                    serial.ReadByte();
                }
                check_timeout = true;
                serial.Write(sendbytes, 0, tx_size_conf);

                conf.configuration_written = true;//write the config to the hardware
            }
            else if (send_now)
            {
                send_now = false;
                unsafe
                {
                    conv.b[0] = 0x00; //checksum start value
                    conv.b[1] = 0x00; //cyclic data

                    conv.b[2] = LoadData.set.mode;
                    conv.b[3] = LoadData.set.ack;
                    conv.f[1] = LoadData.set.curr;
                    conv.f[2] = LoadData.set.power;
                    conv.f[3] = LoadData.set.resistance;
                    conv.f[4] = LoadData.set.voltage;
                    conv.f[5] = LoadData.set.uvlo;
                    //Calculate Checksum

                    
                    for (UInt16 i = 2; i < tx_size; i++)
                    {
                        conv.b[0] ^= conv.b[i];
                    }
                    //Copy the bytes to send them
                    for (UInt16 i = 0; i < tx_size; i++)
                    {
                        sendbytes[i] = conv.b[i];
                    }

                }


                //Throw RX-Buffer away
                recCnt = 0;
                while (serial.BytesToRead > 0)
                {
                    serial.ReadByte();
                }
                check_timeout = true;
                serial.Write(sendbytes, 0, tx_size);

                if (LoadData.set.ack > 0x00)
                {
                    LoadData.set.ack = 0x00;
                }
            }
        }

        private void timer_comm_timeout_Tick(object sender, EventArgs e)
        {

            send_now = true;
            check_timeout = false;
            timer_comm_timeout.Stop();
            recCnt = 0;
            if (serial.IsOpen)
            {
                while (serial.BytesToRead > 0)
                {
                    serial.ReadByte();
                }
            }
            AppendToLog("ERR", "Communication Timeout! Retrying...");
        }

        private void timer_GUI_Tick(object sender, EventArgs e)
        {
            if (rx_ok)
            {
                rx_ok = false;
                txt_act_curr.Text = String.Format("{0:0.0000}", LoadData.act.curr);
                txt_act_power.Text = String.Format("{0:00.000}", LoadData.act.power);
                txt_act_resistance.Text = String.Format("{0:000.00}", LoadData.act.resistance);
                txt_act_voltage.Text = String.Format("{0:00.000}", LoadData.act.voltage);
                txt_act_sinkT.Text = String.Format("{0:00.00}", LoadData.act.temp);
                switch (LoadData.act.mode)
                {
                    case 0:
                        txt_act_mode.Text = "Off";
                        break;
                    case 1:
                        txt_act_mode.Text = "Current";
                        break;
                    case 2:
                        txt_act_mode.Text = "Voltage";
                        break;
                    case 3:
                        txt_act_mode.Text = "Power";
                        break;
                    case 4:
                        txt_act_mode.Text = "Resistance";
                        break;
                    default:
                        txt_act_mode.Text = "???";
                        break;
                }

                txt_act_status.Text = "";
                if (LoadData.act.status == ERR_NO_ERR)
                {
                    txt_act_status.Text = "OK";
                }
                else
                {
                    txt_act_status.Text = "OFF" + Environment.NewLine;
                    if ((LoadData.act.status & ERR_LIM_OVERCURR) > 0)
                    {
                        txt_act_status.Text += "OC" + Environment.NewLine; 
                    }
                    if ((LoadData.act.status & ERR_LIM_OVERVOLT) > 0)
                    {
                        txt_act_status.Text += "OV" + Environment.NewLine;
                    }
                    if ((LoadData.act.status & ERR_LIM_OVERPOW) > 0)
                    {
                        txt_act_status.Text += "OP" + Environment.NewLine;
                    }
                    if ((LoadData.act.status & ERR_LIM_UVLO) > 0)
                    {
                        txt_act_status.Text += "UVLO" + Environment.NewLine;
                    }
                    if ((LoadData.act.status & ERR_LIM_OVERTEMP) > 0)
                    {
                        txt_act_status.Text += "OT" + Environment.NewLine;
                    }
                    if ((LoadData.act.status & ERR_COM_TIMEOUT) > 0)
                    {
                        txt_act_status.Text += "COM TO" + Environment.NewLine;
                    }
                    if ((LoadData.act.status & ERR_COM_CHECKSUM) > 0)
                    {
                        txt_act_status.Text += "COM CHKS" + Environment.NewLine;
                    }
                    if ((LoadData.act.status & ERR_NO_INIT) > 0)
                    {
                        txt_act_status.Text += "INIT" + Environment.NewLine;
                        conf.configuration_written = false;
                        AppendToLog("ERR", "Sending configuration to hardware.");
                    }
                }

                if (cb_logging_enable.Checked)
                {
                    switch (cB_set_Mode.Text)
                    {
                        case "Current":
                            logger.AppendToLog(LoadData.set.curr, "Current", LoadData.act.curr, LoadData.act.voltage, LoadData.act.power, LoadData.act.resistance, LoadData.act.temp);
                            break;
                        case "Power":
                            logger.AppendToLog(LoadData.set.power, "Power", LoadData.act.curr, LoadData.act.voltage, LoadData.act.power, LoadData.act.resistance, LoadData.act.temp);
                            break;
                        case "Resistance":
                            logger.AppendToLog(LoadData.set.resistance, "Resistance", LoadData.act.curr, LoadData.act.voltage, LoadData.act.power, LoadData.act.resistance, LoadData.act.temp);
                            break;
                        case "Voltage":
                            logger.AppendToLog(LoadData.set.voltage, "Voltage", LoadData.act.curr, LoadData.act.voltage, LoadData.act.power, LoadData.act.resistance, LoadData.act.temp);
                            break;
                        default: //Off
                            logger.AppendToLog(0.0f, "Off", LoadData.act.curr, LoadData.act.voltage, LoadData.act.power, LoadData.act.resistance, LoadData.act.temp);
                            break;
                    }
                }
                
            }
            if (check_timeout)
            {
                timer_comm_timeout.Start();
                
            }
            if (checksum_fail)
            {
                checksum_fail = false;
                AppendToLog("ERR", "Communication checksum mismatch!");
                
            }

        }

        private void DataReceivedHandler(object sender, SerialDataReceivedEventArgs e)
        {
            SerialPort sp = (SerialPort)sender;
            byte rec;
            bool err = false;
            unsafe
            {
                while (sp.BytesToRead > 0)
                {
                    rec = (byte)sp.ReadByte();
                    recConv.b[recCnt++] = rec;
                    if (recCnt > bufsize)
                    {
                        err = true;
                        recCnt = 0;
                    }
                }
                if (err == false && recCnt > 0 && recCnt == rx_size)
                {
                    timer_comm_timeout.Stop();
                    //Checksum
                    byte checksum1 = 0x00,checksum2=0xAA;
                    for (UInt16 i = 2; i < recCnt; i++)
                    {
                        checksum1 ^= recConv.b[i];
                        checksum2 ^= recConv.b[i];
                    }
                    if (checksum1 != recConv.b[0] || checksum2 != recConv.b[1])
                    {
                        err = true;
                        checksum_fail = true;
                    }
                        
                }

                if (recCnt == rx_size && recCnt > 0 && err == false)
                {
                    LoadData.act.status = recConv.b[2];
                    LoadData.act.mode = recConv.b[3];
                    LoadData.act.curr = recConv.f[1];
                    LoadData.act.power = recConv.f[2];
                    LoadData.act.resistance = recConv.f[3];
                    LoadData.act.voltage = recConv.f[4];
                    LoadData.act.temp = recConv.f[5];
                    send_now = true;
                    rx_ok = true;
                }
              
            }
           
        }

        public void AppendToLog(string type, string msg)
        {
            txt_log.AppendText( "[" + type + "] (" + System.DateTime.Now.ToString() + ") - " + msg + Environment.NewLine);
            txt_log.Select(txt_log.Text.Length, 0);
            txt_log.ScrollToCaret();
        }

        private void mainWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            bool close = true;
            if (logger.GetMeasLines() > 0)
            {
                if (MessageBox.Show("Do you really want to close the application?" + Environment.NewLine + "Unsafed log files are lost.", project_name + project_version,
              MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No)
                {
                    close = false;
                    e.Cancel = true;
                    
                }
            }
            if(close)
            {
                if (thread_active)
                {
                    thread_active = false;
                    comm_thread.Abort();
                }
                
            }
            
        }


        //Hardware config
        const string conf_id_max_power                  = "MAX_POWER";
        const string conf_id_max_voltage                = "MAX_VOLTAGE";
        const string conf_id_max_current                = "MAX_CURRENT";
        const string conf_id_max_temperature            = "MAX_TEMPERATURE";
        const string conf_id_scale_curr_in              = "INPUT_CURRENT_SCALING";
        const string conf_id_scale_curr_out             = "OUTPUT_CURRENT_SCALING";
        const string conf_id_scale_volt_in              = "INPUT_VOLTAGE_SCALING";
        const string conf_id_current_controller_kI      = "CURRENT_CONTROLLER_KI";
        const string conf_id_current_controller_limit   = "CURRENT_CONTROLLER_LIMIT";
        const string conf_id_voltage_controller_kP      = "VOLTAGE_CONTROLLER_KP";
        const string conf_id_voltage_controller_kI      = "VOLTAGE_CONTROLLER_KI";
        const string conf_id_cable_resistance           = "CABLE_RESISTANCE";

        //GUI config
        const string conf_id_std_baud       = "COMM_BAUD_RATE";
        const string conf_id_std_cycleT     = "COMM_CYCLE_TIME_MS";

        private void InitConfigData()
        {
            conf.max_current = 3.0f;
            conf.max_power = 90.0f;
            conf.max_voltage = 30.0f;
            conf.max_temperature = 60.0f;
            conf.current_input_scale = 0.00124214f;
            conf.current_output_scale = 774.381869f;
            conf.voltage_input_scale = 0.01306865f;
            conf.cable_resistance = 0.0f;

            conf.current_controller_kI = 1.0f;
            conf.current_controller_limit = 0.02f;

            conf.voltage_controller_kP = 0.1f;
            conf.voltage_controller_kI = 1.0f;

            conf.configuration_loaded = false;
            conf.configuration_written = false;

        }

        private void ReadConfigData()
        {
            ConfigFileReader confReader;
            string path = System.IO.Directory.GetCurrentDirectory() + "\\" + "ELOAD.config";
            confReader = new ConfigFileReader(path);
            confReader.ReadConfigFile();


            conf.max_current = confReader.GetFloatFromFile(conf_id_max_current);
            conf.max_power = confReader.GetFloatFromFile(conf_id_max_power);
            conf.max_voltage = confReader.GetFloatFromFile(conf_id_max_voltage);
            conf.max_temperature = confReader.GetFloatFromFile(conf_id_max_temperature);
            conf.current_input_scale = confReader.GetFloatFromFile(conf_id_scale_curr_in);
            conf.current_output_scale = confReader.GetFloatFromFile(conf_id_scale_curr_out);
            conf.voltage_input_scale = confReader.GetFloatFromFile(conf_id_scale_volt_in);
            conf.current_controller_kI = confReader.GetFloatFromFile(conf_id_current_controller_kI);
            conf.current_controller_limit = confReader.GetFloatFromFile(conf_id_current_controller_limit);

            conf.voltage_controller_kP = confReader.GetFloatFromFile(conf_id_voltage_controller_kP);
            conf.voltage_controller_kI = confReader.GetFloatFromFile(conf_id_voltage_controller_kI);

            conf.cable_resistance = confReader.GetFloatFromFile(conf_id_cable_resistance);

            txt_comm_baud.Text = confReader.GetIntFromFile(conf_id_std_baud).ToString();
            txt_comm_cycle.Text = confReader.GetIntFromFile(conf_id_std_cycleT).ToString();

            if (confReader.HasFailedAtLeastOnce() || confReader.IdentifierExistsInFile("CONF_SAMPLE_CONFIG"))
            {
                //Oh no
                conf.configuration_loaded = false;
            }
            else
            {
                conf.configuration_loaded = true;
            }

        }

        private void WriteStdConfigData()
        {
            
            ConfigFileWriter confWriter;
            string path = System.IO.Directory.GetCurrentDirectory() + "\\" + "ELOAD.config";

            conf.max_voltage = 30.0f;
            conf.max_current = 1.0f;
            conf.max_power = 10.0f;

            confWriter = new ConfigFileWriter(path);
            confWriter.AddCommentLine("This file was auto generated. Change the values according to your limits.");
            confWriter.AddFloatToFile("CONF_SAMPLE_CONFIG", 1.0f, "This identifier indicates that your config is incorrect. Remove this identifier if config is valid.");
            confWriter.AddFloatToFile(conf_id_max_power, conf.max_power, "Maximum Power in W");
            confWriter.AddFloatToFile(conf_id_max_voltage, conf.max_voltage, "Maximum Voltage in V");
            confWriter.AddFloatToFile(conf_id_max_current, conf.max_current, "Maximum Current in A");
            confWriter.AddFloatToFile(conf_id_max_temperature, conf.max_temperature, "Maximum Sink temperaure in degC");
            confWriter.AddFloatToFile(conf_id_scale_curr_in, conf.current_input_scale, "Current measurement scaling in A / LSB");
            confWriter.AddFloatToFile(conf_id_scale_curr_out, conf.current_output_scale, "Current output scaling in LSB / A");
            confWriter.AddFloatToFile(conf_id_scale_volt_in, conf.voltage_input_scale, "Voltage measurement scaling in Volts / LSB");

            confWriter.AddFloatToFile(conf_id_current_controller_kI, conf.current_controller_kI, "Current control kI");
            confWriter.AddFloatToFile(conf_id_current_controller_limit, conf.current_controller_limit, "Current control limit in A");
            confWriter.AddFloatToFile(conf_id_voltage_controller_kP, conf.voltage_controller_kP, "Voltage control kP");
            confWriter.AddFloatToFile(conf_id_voltage_controller_kI, conf.voltage_controller_kI, "Voltage control kI");
            confWriter.AddFloatToFile(conf_id_cable_resistance, conf.cable_resistance, "Cable resistance in Ohms");


            confWriter.AddIntToFile(conf_id_std_baud, 115200,"Standard startup Baudrate in 1/s");
            confWriter.AddIntToFile(conf_id_std_cycleT, 500,"Standard Cycle time in ms");
            confWriter.WriteConfigFile();
        }

        private void ConfigureLoadDataStorage()
        {
            LoadData.set.curr = 0.0f;
            LoadData.set.voltage = 0.0f;
            LoadData.set.power = 0.0f;
            LoadData.set.uvlo = 5.0f;
            LoadData.set.resistance = 1000.0f;
            LoadData.set.mode = 0;

            LoadData.act.curr = 0.0f;
            LoadData.act.voltage = 0.0f;
            LoadData.act.power = 0.0f;
            LoadData.act.resistance = 0.0f;
            LoadData.act.status = 0;
            LoadData.act.mode = 0;
            
        }

    }
}
